from poisoning_adversary import get_connections, unzip
import poisoned_datasets as pd
import numpy as np
from typing import List, Callable, Union
import os
from tqdm import tqdm
import torchvision.datasets as tds
from argparse import ArgumentParser

'''
This module provides one main method: search_degree_gamma, which takes in
points, a function that selectes degrees from a list of several degrees, and
returns a gamma. The degree_func should be such that binary search is possible,
which this method makes use of.
'''

GAMMA_FUNCTIONS = {
    'max': max,
    'average': lambda x: sum(x)/len(x)
}

def search_degree_gamma(
    points,
    degree_limit: int, 
    lower_limit: float, 
    upper_limit: float, 
    iterations: int,
    norm: float,
    batch_size: int,
    degree_func: Callable[[List[int]], Union[int, float]],
    use_tqdm: bool = True) -> List[List[int]]:
    '''
    Finds out the largest gamma such that the graph produced has nodes with max
    degree <= degree_limit
    '''
    if len(points) == 0:
        return None, None
    avg_degree = None
    iterable = tqdm(range(iterations)) if use_tqdm else range(iterations)
    for iteration in iterable:
        mid = (lower_limit + upper_limit) / 2.0
        adjacency_list = get_connections(
            points=points,
            norm=norm,
            norm_threshold=mid,
            batch_size=batch_size
        )
        degrees = [len(adj) for adj in adjacency_list]
        average_degree = degree_func(degrees)
        if average_degree > degree_limit:
            upper_limit = mid
        else:
            lower_limit = mid
            avg_degree = average_degree
    
    # Remove this
    # adjacency_list = get_connections(
    #     points=points,
    #     norm=norm,
    #     norm_threshold=lower_limit,
    #     batch_size=batch_size
    # )
    # adjacency_list = [len(l) for l in adjacency_list]
    # from collections import defaultdict
    # degrees = defaultdict(lambda: 0)
    # for degree in adjacency_list:
    #     degrees[degree] += 1
    # print(sorted(list(degrees.items())))
    # Remove before this
    return (lower_limit, avg_degree)


'''
The following functions are useful only for analysis, i.e., running this file as
__main__
'''

def get_gammas(dataset, gamma_getter):
    '''
    Produces gammas for each class in dataset 
    gamma_getter is a function that takes in points of a particular class, and
        returns a gamma value for that class
    '''
    num_classes = len(set(unzip(dataset)[1]))
    classwise_points = [[] for _ in range(num_classes)]
    for x, y in dataset:
        classwise_points[y].append(np.asarray(x).astype(np.float32)/255)
    gammas = [gamma_getter(points) for points in classwise_points]
    return gammas


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--dataset', type=str)
    args = parser.parse_args()
    # dataset = get_dataset_by_name(args.dataset)
    dataset = pd.DATASETS[args.dataset](os.path.join(os.environ['HOME'], 'datasets/cifar')).train_set
    def gamma_getter(points):
        return search_degree_gamma(
            points=points,
            degree_limit=30,
            lower_limit=0.0,
            upper_limit=0.8,
            iterations=10,
            norm=float('inf'),
            batch_size=128,
            degree_func=lambda x: max(x),
            use_tqdm=True
        )
    gammas = get_gammas(dataset=dataset, gamma_getter=gamma_getter)
    for i, (x, y) in enumerate(gammas):
        print('class {}, gamma: {}, degree: {}'.format(i, x, y))
